/* @pjs crisp=true; 
preload="sea.jpg"; 
 */

PImage img, original;
int pixels_length = 0;
int histogram_channel = 0;

class Histogram {
  int[] r, g, b, intensity;
  int max_r = 0, max_g = 0, max_b = 0, max_i = 0;
  
  Histogram() {
    r = new int[256];
    g = new int[256];
    b = new int[256];
    intensity = new int[256];
    
    clear();
  }
  
  void add(int r, int g, int b) {
    this.r[r]++;
    this.g[g]++;
    this.b[b]++;
    
    if (this.r[r] > max_r) {
      max_r = this.r[r];
    }
    
    if (this.g[g] > max_g) {
      max_g = this.g[g];
    }
    
    if (this.b[b] > max_b) {
      max_b = this.b[b];
    }
    
    int i = (r+g+b)/3;
    this.intensity[i]++;
    if (this.intensity[i] > max_i) {
      max_i = this.intensity[i];
    }
  }
  
  void clear() {
    for (int i = 0; i < 256; i++) {
      r[i] = g[i] = b[i] = intensity[i] = 0;
    }
    max_r = max_g = max_b = max_i = 0;
  }
}

Histogram histogram = new Histogram();

boolean show_original = true;

void copyImage() {
  img.copy(original, 0, 0, original.width, original.height, 0, 0, original.width, original.height);
}

void setup() {
  size(640, 360);
  noFill();
  stroke(255);
  frameRate(30);
  original = loadImage("sea.jpg");
  img = createImage(original.width, original.height, RGB);
  pixels_length = original.pixels.length;
  copyImage();
  if (show_original) {
    pixels_length /= 2;
  }
  
  posterize(img, 2);
}

void horizontalMosaic(PImage source, int bloc_length) {
  if (bloc_length%3 == 0)
    bloc_length++;
  source.loadPixels();
  for (int i = 0; i < pixels_length; i += bloc_length) {
    source.pixels[i] &= 0xff*((int)pow(256, (i%3)));
    
    for (int j = 1; j < bloc_length; j++) {
      source.pixels[i+j] = source.pixels[i];
    }
  }
  source.updatePixels();
}

void colorStore(PImage source, int lines) {
  int bloc_length = source.width*lines;
  
  source.loadPixels();
  for (int i = 0; i < pixels_length-bloc_length; i += bloc_length) {
    for (int j = 0; j < bloc_length; j++) {
      int index = i+j;
      // << 3 -> * 8
      source.pixels[index] = source.pixels[index]&0xff<<((i%3) << 3);
    }
  }
  source.updatePixels();
}

void toGreyscale(PImage source, float luminosity) {
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int r = source.pixels[i]>>16&0xff;
    int g = source.pixels[i]>>8&0xff;
    int b = source.pixels[i]&0xff;
    int intensity = (int)((r+g+b)/luminosity);
    if (intensity > 255) {
      intensity = 255;
    }
    source.pixels[i] = intensity<<16 |
                       intensity<<8 |
                       intensity;
  }
  source.updatePixels();
}

void applyFunction(PImage source, float level, int notchRange) {
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int r = source.pixels[i]>>16&0xff;
    int g = source.pixels[i]>>8&0xff;
    int b = source.pixels[i]&0xff;
    int mean = (r+g+b)/3;
    if (mean < notchRange && mean > 255-notchRange) {
    /*r = (int)(1/(log(r)/log(level)))*255;
    g = (int)(1/(log(g)/log(level)))*255;
    b = (int)(1/(log(b)/log(level)))*255;
    */
    
   /* r = (int)(cos(level)*r);
    g = (int)(sin(level)*g);
    b = (int)(tan(level)*b);
    */
    
    r = (int)pow(r, level);
    g = (int)pow(g, level);
    b = (int)pow(b, level);
    
    r = constrain(r, 0, 255);
    g = constrain(g, 0, 255);
    b = constrain(b, 0, 255);
    
    /*r = 255-r;
    g = 255-g;
    b = 255-b;*/
    }
    source.pixels[i] = r<<16 |
                       g<<8 |
                       b;
  }
  source.updatePixels();
}

void posterize(PImage source, int level) {
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int r = source.pixels[i]>>16&0xff;
    int g = source.pixels[i]>>8&0xff;
    int b = source.pixels[i]&0xff;
    
    int lr = 255-r;
    int lg = 255-g;
    int lb = 255-b;
   
    r = round(r/level)*level;
    g = round(g/level)*level;
    b = round(b/level)*level;
   
    /*
    r+= round(lr/level);
    g+= round(lg/level);
    b+= round(lb/level);
    */
    source.pixels[i] = r<<16 |
                       g<<8 |
                       b;
  }
  source.updatePixels();
}

void scroll(PImage source) {
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int index = (floor(i/source.width)*source.width);
    int value = source.pixels[index];
    source.pixels[index] = source.pixels[i];
    source.pixels[i] = value;
  }
  source.updatePixels();
}

void flip(PImage source) {
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int flipped_index = pixels_length-i%1;
    int value = source.pixels[flipped_index];
    source.pixels[flipped_index] = source.pixels[i];
    source.pixels[i] = value;
  }
  source.updatePixels();
}

void applyNoise(PImage source, int type, int seed, float amount, int intensity) {
  randomSeed(seed);
  
  int intensity_d2 = intensity/2;
  
  source.loadPixels();
  for (int i = 0; i < pixels_length; i++) {
    int r = source.pixels[i]>>16&0xff;
    int g = source.pixels[i]>>8&0xff;
    int b = source.pixels[i]&0xff;
    
    if (random(1.0) > amount) {
      int n = 0;
      if (type == 0) // uniform
      {
        n = (int)(intensity_d2-random(intensity));
        r += n;
        g += n;
        b += n;
      }
      else // salt&pepper
      {
        n = (int)round((random(255)/255))*255;
        r = n;
        g = n;
        b = n;
      }
    }
    
    if (r > 255) r = 255;
    if (g > 255) g = 255;
    if (b > 255) b = 255;
    if (r < 0) r = 0;
    if (g < 0) g = 0;
    if (b < 0) b = 0;
    
    source.pixels[i] = r<<16 | g<<8 | b;;
  }
  source.updatePixels();
}

void convolution(PImage source, int filter_width, int filter_height) {
  int image_height = pixels_length/source.width;
  
  int filter_size = filter_width * filter_height;
  
  int[] rr = new int[filter_size];
  int[] gg = new int[filter_size];
  int[] bb = new int[filter_size];
  int filter_sized2 = floor(filter_size/2);
      
  source.loadPixels();
  for (int x = 0; x < source.width; x++) {
    for (int y = 0; y < image_height; y++) {
      int rm = 0, gm = 0, bm = 0;
      
      for (int i = 0; i < filter_width; i++) {
        for (int j = 0; j < filter_height; j++) { 
          int index = ((x - filter_width / 2 + i + source.width)%source.width)+((y - filter_height / 2 + j + image_height) % image_height)*source.width;
          int r = source.pixels[index]>>16&0xff;
          int g = source.pixels[index]>>8&0xff;
          int b = source.pixels[index]&0xff;
          
          rm += r;
          gm += g;
          bm += b;
          
          int f_index = i+j*filter_width;
          rr[f_index] = r;
          gg[f_index] = g;
          bb[f_index] = b;
        }
      }
      /*
      rm /= filter_size;
      gm /= filter_size;
      bm /= filter_size;
      */
      
      rr = sort(rr);
      gg = sort(gg);
      bb = sort(bb);
      
      rm = rr[filter_sized2];
      gm = gg[filter_sized2];
      bm = bb[filter_sized2];
      
      int index = x+y*source.width;
      source.pixels[index] = rm<<16 | gm<<8 | bm;
    }
  }
  source.updatePixels();
}

void non_local_filter(PImage source, int filter_width, int filter_height, int max_search, int search_window_size) {
  PImage source_clone = source.get();
  int image_height = pixels_length/source.width;
  
  int filter_size = filter_width * filter_height;
  
  int filter_sized2 = floor(filter_size/2);

  source.loadPixels();
  source_clone.loadPixels();
  for (int x = 0; x < source.width; x++) {
    for (int y = 0; y < image_height; y++) {
      int rm = 0, gm = 0, bm = 0;
      int main_index = x+y*source.width;
      
      for (int i = 0; i < filter_width; i++) {
        for (int j = 0; j < filter_height; j++) { 
          int index = ((x - filter_width / 2 + i + source.width)%source.width)+((y - filter_height / 2 + j + image_height) % image_height)*source.width;
          int r = source_clone.pixels[index]>>16&0xff;
          int g = source_clone.pixels[index]>>8&0xff;
          int b = source_clone.pixels[index]&0xff;
          
          rm += r;
          gm += g;
          bm += b;
        }
      }
      

  
      rm /= filter_size;
      gm /= filter_size;
      bm /= filter_size;
      
      //int om = (rm+gm+bm)/3;
      
        int mr = 32;
        int mg = 32;
        int mb = 32;
      for (int s = 0; s < max_search; s++) {
        int xx = x + search_window_size-(int)random(search_window_size+search_window_size);
        int yy = y + search_window_size-(int)random(search_window_size+search_window_size);
        if (xx > source.width) xx = source.width-1;
        if (yy > image_height) yy = image_height-1;
        if (xx < 0) xx = 0;
        if (yy < 0) yy = 0;
        int tnrr = 0, tngg = 0, tnbb = 0;
        int ind = xx+yy*source.width;
        int tr = source_clone.pixels[ind]>>16&0xff;
        int tg = source_clone.pixels[ind]>>8&0xff;
        int tb = source_clone.pixels[ind]&0xff;
           
        for (int i = 0; i < filter_width; i++) {
          for (int j = 0; j < filter_height; j++) { 
            int index = ((xx - filter_width / 2 + i + source.width)%source.width)+((yy - filter_height / 2 + j + image_height) % image_height)*source.width;
            int r = source_clone.pixels[index]>>16&0xff;
            int g = source_clone.pixels[index]>>8&0xff;
            int b = source_clone.pixels[index]&0xff;
            
            tnrr += r;
            tngg += g;
            tnbb += b;
          }
        }
        
        tnrr /= filter_size;
        tngg /= filter_size;
        tnbb /= filter_size;
    
        if (abs(rm-tnrr) < mr) {
          if (abs(gm-tngg) < mg) {
            if (abs(bm-tnbb) < mb) {
        
         mr = abs(rm-tnrr);
         mg = abs(gm-tngg);
         mb = abs(bm-tnbb);

         rm = tr;
         gm = tg;
         bm = tb;
 
            }
          }
        }
      }

      source.pixels[main_index] = rm<<16 | gm<<8 | bm;
    }
  }
  source.updatePixels();
}

void buildHistogram(PImage source, int precision) {
  histogram.clear();
  
  source.loadPixels();
  for (int i = 0; i < pixels_length; i+=precision) {
    int r = source.pixels[i]>>16&0xff;
    int g = source.pixels[i]>>8&0xff;
    int b = source.pixels[i]&0xff;
    
    histogram.add(r, g, b);
  }
}

void displayHistogram(int opacity, int histoHeight, int histoChannel) {
  noStroke();
  int middle = width/2-256;
  
  int opacity_r = 0;
  int opacity_g = 0;
  int opacity_b = 0;
  int opacity_i = 0;
  
  if (histoChannel == 0 || histoChannel == 1) {
    opacity_r = opacity;
  } 
  if (histoChannel == 0 || histoChannel == 2) {
    opacity_g = opacity;   
  }
  if (histoChannel == 0 || histoChannel == 3) {
    opacity_b = opacity;   
  }
  if (histoChannel == 4) {
    opacity_i = opacity;   
  }
  
  for (int i = 0; i < 256; i++) {
    fill(255, 0, 0, opacity_r);
    rect(i*2+middle, height, 3, -(round((float)histogram.r[i]/(float)histogram.max_r*histoHeight)));
    fill(0, 255, 0, opacity_g);
    rect(i*2+middle, height, 3, -(round((float)histogram.g[i]/(float)histogram.max_g*histoHeight)));
    fill(0, 0, 255, opacity_b);
    rect(i*2+middle, height, 3, -(round((float)histogram.b[i]/(float)histogram.max_b*histoHeight)));
    fill(255, 255, 255, opacity_i);
    rect(i*2+middle, height, 3, -(round((float)histogram.intensity[i]/(float)histogram.max_i*histoHeight)));
  }
  noFill();
}

int effect = 0;
int scroll_timer = millis();

String effectName = "posterize";
void draw() {
  background(0);
  
  copyImage();
  if (effect == 0) {
    posterize(img, (int)map(mouseX, 0, width, 1, 255));
    effectName = "posterize";
  } else if (effect == 1){
    toGreyscale(img, map(mouseX, 0, width, 0.0001, 2.0));
    effectName = "greyscale";
  } else if (effect == 2) {
    colorStore(img, (int)map(mouseX, 0, width, 1, 3));
    effectName = "store";
  } else if (effect == 3) {
    horizontalMosaic(img, (int)map(mouseX, 0, width, 1, 32));
    effectName = "horizontal mosaic";
  } else if (effect == 4) {
    applyFunction(img, map(mouseX, 0, width, 0.5, 1.5), (int)map(mouseY, 0, height, 0, 255));
    effectName = "function (notch filter)";
  } else if (effect == 5) {
    applyNoise(img, 0, 0, 0.5, 255);
    convolution(img, 3, 3);
    effectName = "convolution (denoise)";
  } else if (effect == 6) {
    applyNoise(img, 0, 0, 0.5, 255);
    non_local_filter(img, 3, 3, 4, 32);
    effectName = "non local mean filter";
  }
  
  if (millis() - scroll_timer > 2) {
    scroll(original);
    scroll_timer = millis();
  }
  
  image(img, 0, 0);
  
  buildHistogram(img, 16);
  displayHistogram((int)map(mouseY, height, 0, 64, 255), height-img.height/2, histogram_channel);
  
  fill(255);
  textSize(32);
  text(effectName,2,height-6);
}

void mouseReleased() {
  effect++;
  effect%=7;
}

void keyPressed() {
  if (keyCode == DOWN) {
    applyNoise(img, 0, 0, 2, 255);
    image(img, 0, 0);
  }
}

void keyReleased() {
  if (keyCode == RIGHT) {
    histogram_channel++;
    histogram_channel%=5;
  } else if(keyCode == LEFT) {
    histogram_channel--;
    if (histogram_channel < 0) {
      histogram_channel = 4;
    }
  } else if(keyCode == UP) {
    flip(original);
  }
}
